import numpy as np
import pandas as pd
import os
import collections
import warnings
from sklearn.metrics import mean_squared_error, mean_absolute_error
from sklearn.metrics import r2_score
from numpy import linalg as LA
from itertools import product
from matplotlib import pyplot as plt

from scipy.stats import pearsonr
from sklearn.decomposition import PCA

# Try to load seaborn 
# (not necessary but makes plot look nicer)
try:
    import seaborn as sns
    sns.set_style("ticks")
except ImportError:
    pass


def preprocess(X, method):
    """
    Data preprocessing.
    
    ----- Parameters
    X - data
    method - method to be used
    ----- Reterns
    ne data
    """

    from sklearn import preprocessing
# does not work for PLS; works for linear and SVR and GTB
    if (method == 'mean'):
        Xnew = preprocessing.StandardScaler(copy=True, with_mean=True, with_std=True).fit_transform(X) # zero mean and unit variance
    elif (method == 'minmax'):
        Xnew = preprocessing.MinMaxScaler(copy=True, feature_range=(0, 1)).fit_transform(X)
#--------------- RobustScaler does not work - consider everything as outliers ?
    #scaler = preprocessing.RobustScaler(copy=True).fit_transform(X)
#-----
    #scaler = preprocessing.QuantileTransformer(copy=True,n_quantiles=5, random_state=0).fit_transform(X) # nonlinear
    #scaler = preprocessing.PowerTransformer(copy=True,method='box-cox', standardize=False).fit_transform(X)
#---------------------
#---- bad for Ridge and PLS ------------
#    Xnew =  preprocessing.scale(X,axis=1)  #zero mean and unit variance for each feature 
#------- here PLS works well, SVR does not really work
#    Xnew =  preprocessing.normalize(X,axis=1)
#    Xnew = X
    return(Xnew)


#################################################################
#  Visualization for ML
#################################################################
def plot_GridSearchCV(clf,Split,alphas = []):
    """
    Plot  cross validation score for the set of hyperparameters
    
    ----- Parameters
    clf - output object of the GridSearchCV hyperparameters scan function
    Split - number of training/validation set splitting used in GridSearchCV 
    ----- Reterns
    alphas - array of the values of hyperparameters used for optimization in GridSearchCV
    """
    from sklearn.model_selection import GridSearchCV
    scores = clf.cv_results_['mean_test_score']
    scores_std = clf.cv_results_['std_test_score']
    if(len(alphas) == 0):
        alphas = range(0,len(scores))
    plt.figure().set_size_inches(4, 3)
    plt.semilogx(alphas, scores)
    # plot error lines showing +/- std. errors of the scores
    std_error = scores_std / np.sqrt(Split)

    plt.semilogx(alphas, scores + std_error, 'b--')
    plt.semilogx(alphas, scores - std_error, 'b--')

    # alpha=0.2 controls the translucency of the fill color
    plt.fill_between(alphas, scores + std_error, scores - std_error, alpha=0.2)
   
    plt.ylabel('CV score +/- std error')
    plt.xlabel('alpha')
    plt.axhline(np.max(scores), linestyle='--', color='.5')
    plt.xlim([alphas[0], alphas[-1]])
    plt.ylim(-0.5,1)
    plt.show()

    return

def plot_fitting_train_test_ramd(clf,X,y,y_err,ramd,ramd_err,k_fold,Split,dat='koff'):
    """
    Plot train/validation results of RMs
    
    ----- Parameters
    clf - RM  object 
    X - data
    y- labels
    y_err - SD of labels
    ramd - tauRAMD simulation data
    ramd_err - SD of tauRAMD data
    k_fold - object with splitted train/validation sets
    Split - number of splitting
    dat - type of rate-constant
    ----- Reterns
    scores of the method evaluation -    r2_train,mae_train,mae_test,mae_ramd
    coef - coefficients of the LR model
    """
    from sklearn.model_selection import ShuffleSplit
    fig = plt.figure(figsize = (16, 8),facecolor='w')
    fig.subplots_adjust(hspace=0.3, wspace=0.3)
    sns.set(font_scale=1.4)
    sns.set_style("ticks")
    if dat =='koff':
        xlim1,xlim2 = -2.5,4.5
        ylim1,ylim2 = -2.5, 4.5
    if dat =='kd':
        xlim1,xlim2 = 5,12
        ylim1,ylim2 = 2,13
    if dat =='kon':
        xlim1,xlim2 = 3,7
        ylim1,ylim2 = 2.5,7

    index = 1
    mae_train = []
    mae_test = []
    mae_ramd = []
    r2_train = []
    coef = []
    plot_part = 8
    for train_index, test_index in k_fold:
        X_train, X_test = X[train_index], X[test_index]
        y_train, y_test = y[train_index], y[test_index] 
        ramd_train, ramd_test = ramd[train_index], ramd[test_index] 
        ramd_err_train, ramd_err_test = ramd_err[train_index], ramd_err[test_index] 
        y_err_train,y_err_test = y_err[train_index],y_err[test_index]
        clf.fit(X_train, y_train)
        y_pred_train = clf.predict(X_train)
        y_pred_test = clf.predict(X_test)
        r2_train.append(r2_score(y_train, y_pred_train))
        mae_train.append(mean_absolute_error(y_train, y_pred_train))
        mae_test.append(mean_absolute_error(y_test, y_pred_test))
        mae_ramd.append(mean_absolute_error(ramd_test,y_test))
        if(index <= plot_part):
            ax = plt.subplot(plot_part/4,4,index)
            plt.xlim(xlim1,xlim2)
            plt.ylim(ylim1,ylim2)
            ax.plot(np.linspace(xlim1-1,xlim2+1, 20),np.linspace(ylim1-1,ylim2+1, 20),c = 'k')
            ax.errorbar(y_train,y_pred_train,xerr = y_err_train,fmt = 'o', c = 'royalblue', alpha = 0.5, markersize=8)
            ax.errorbar(y_test,ramd_test,yerr = ramd_err_test,fmt = 'o', c = 'green',  alpha = 0.5,markersize=8)
            ax.scatter(y_test, y_pred_test, marker = 'o', facecolor = 'darkorange',s=50) 
#        ax.scatter(y_test, y_D_test, marker = 'o', facecolor = 'red',s=60) 
            ax.set_ylabel("predicted tau [s] ", fontsize = 14)
            ax.set_xlabel("log10(1/"+dat+") [s] ", fontsize = 14)
        index += 1
    plt.show()
    
    fig = plt.figure(figsize = (16, 4),facecolor='w')
    plt.subplot(1,3,1)
    plt.xlim(0.5,1)
    ax=sns.distplot(r2_train, hist=True, kde=True, norm_hist = True,
             hist_kws={'edgecolor':'black'}, color='royalblue',kde_kws={'linewidth': 4})        
    ax.set_ylabel("R2 train", fontsize = 14)
    ax.set_xlabel("log10(1/"+dat+") [s] ", fontsize = 14)
    plt.subplot(1,3,2)
    plt.xlim(0,0.6)
    ax=sns.distplot(mae_train, hist=True, kde=True, norm_hist = True,
             hist_kws={'edgecolor':'black'}, color='royalblue',kde_kws={'linewidth': 4})        
    ax.set_ylabel("MAE train", fontsize = 14)
    ax.set_xlabel("log10(1/"+dat+") [s] ", fontsize = 14)
    plt.subplot(1,3,3)
    plt.xlim(0,2)
    ax=sns.distplot(mae_test, hist=True, kde=True, norm_hist = True,label="TEST",
             hist_kws={'edgecolor':'black'}, color='darkorange',kde_kws={'linewidth': 4})
    ax=sns.distplot(mae_ramd, hist=False, kde=True,norm_hist = True, bins = 5,label="tauRAMD",
             hist_kws={'edgecolor':'black'}, color='green',kde_kws={'linewidth': 4})
    ax.set_ylabel("MAE ", fontsize = 14)
    ax.set_xlabel("log10(1/"+dat+") [s] ", fontsize = 14)
    plt.legend(loc='upper right',framealpha = 0.0,edgecolor="None")
    plt.show()
    
    return (r2_train,mae_train,mae_test,mae_ramd,coef)


################################################################
# Visualization fo any clustering methods  - depricated
#
################################################################
def any_cluster_plot(average_data_comp,average_comp,labels,cluster_centers,par='koff',feature=16,features=[]):
    """
    Visualization fo any clustering methods  
    
    ----- Parameters
    average_data_comp - pandas dataframe
    average_comp - numpy matrix of the data (essentially the same as  average_data_comp but containing only features)
    labels - array of claster lables (which compound belong to which cluster)
    cluster_centers - mean weights of features for each claster
    feature - number of features 
    features - list of feature name
    ----- Reterns
    nothing
    """
    
    n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
    core_samples_mask = np.zeros_like(labels, dtype=bool)
    unique_labels = set(labels)
    colors = [plt.cm.Spectral(each)
          for each in np.linspace(0, 1, len(unique_labels))]
    sns.set_style("ticks")
    sns.set(font_scale=1.1)
    fig=plt.figure(figsize = (16, 8),facecolor ='w')

    plt.title('Estimated number of clusters: %d' % n_clusters_)
    
    i_sorted = np.argsort(np.asarray(average_data_comp['ramd']), axis=0)

# we will rename compound types to plot them
#    average_data_comp1 = np.zeros(len(average_data_comp['type'].tolist()))
#    print(np.unique(average_data_comp1))
#    for i, ty in enumerate(np.unique(np.asarray(average_data_comp['type']))):
#        for j,zz in enumerate(np.asarray(average_data_comp['type'])):
#            if zz == ty:
#                average_data_comp1[j] = i
    ccat = average_data_comp['group'].unique()
    average_data_comp1=average_data_comp['group'].astype("category").cat.rename_categories(np.linspace(0, len(ccat)-1, len(ccat)))

    for k, col in zip(unique_labels, colors):
        class_member_mask = (labels == k)
        xy = average_comp[class_member_mask]
        y = average_comp[class_member_mask]
        x1 = average_data_comp[par][class_member_mask]
        z = np.asarray(average_data_comp['group'][class_member_mask])
        w = np.asarray(average_data_comp1)[class_member_mask]
        comp =  average_data_comp.index[class_member_mask]
#        print(k,len(comp),max(x1)-min(x1),comp.tolist(),z.tolist())
#-----------------------PLOT 1 - 1---------------------
        plt.subplot(132)
        if(par == 'ramd'):   plt.xlabel('log10('+par+' [ns])')
        else:     plt.xlabel('log10(1/'+par+' [s])')            
        plt.ylabel('group')
        plt.yticks(np.arange(len(np.unique(np.asarray(average_data_comp['group'])))), np.unique(np.asarray(average_data_comp['group'])))
        plt.title(average_data_comp.columns.tolist()[feature])    
        plt.plot(x1, w, 'o', markerfacecolor=tuple(col),markeredgecolor='k', markersize=12, markeredgewidth=1.)
        plt.legend(unique_labels,loc='upper right')
#-------------------------PLOT 1 - 2-------------------------------        
        plt.subplot(133)
        plt.title(average_data_comp.columns.tolist()[feature])    
        if(par == 'ramd'):   plt.xlabel('log10('+par+' [ns])')
        else:     plt.xlabel('log10(1/'+par+' [s])')
        plt.plot(x1, y[:, feature], 'o', markerfacecolor=tuple(col), markeredgecolor='k', markeredgewidth=1., markersize=14)
#-------------------------PLOT 1 - 3-----------------------------        
    plt.subplot(131)
    if(len(features) > 1):
        plt.yticks(np.arange(len(features)), features)
        plt.xticks(np.arange(len(cluster_centers)), np.arange(len(cluster_centers)))
        anorm =normalize(cluster_centers, axis=1)
        plt.imshow(anorm.T,cmap= 'Blues',aspect ="auto")
    else:
        plt.imshow(cluster_centers.T,cmap= 'Blues',aspect ="auto")
    plt.xlabel('cluster')
    plt.ylabel('feature')
    plt.subplots_adjust(left=0, wspace=0.1, top=0.8)    
    plt.show()
  
    
    return
