
import numpy as np
import os
import csv
from collections import OrderedDict
import re
import sys

###########################################################################
# Helper function for correcting the wrong numbering that some ligands have
# (Ask Daria for information on why this is necessary)
###########################################################################

def correct_single_numbering(feature):
    '''
    Corrects wrong numbering of residues (by subtracting 1 from the number)
    Is called on single features.

    Example:
        Input:  'APO_X14ALA'
        Output: 'APO_X13ALA'
    '''
    number = re.findall(r"\d+", feature)
    if len(number):
        number = int(number[0])
        feature = re.sub(str(number), str(number - 1), feature)

    return feature


#############################################################
# Data loading routines
# These functions take the raw data and (after preprocessing) 
# save them as numpy files.
#############################################################



def load_full_data(
                root_dir = None, 
                ligands = [],
                out_dir = None,
                data_folder = None,
                interaction_filename = "full-interaction-profile.tsv", 
                initiation_threshold = 2, 
                print_missing_data = False):
    """Loading the raw data stored in 'full-interaction-profile.tsv'.
    First a list of all subdirectories of `out_dir` is created. Every 
    subdirectory corresponds to a single trajectory.
    Next, by looping through all subdirectories, a list of all features is 
    created (this is necessary as not all trajectories have the same 
    features) as well as a list of all hyperparameters (i.e. ligand, starting
    position and number of trajectory).
    In this step, the numbering of features of the ligands specified in 
    `wrong_numbering` is corrected (this is due to some problems with the 
    original data constructed by Daria.)
    An empty array called `data` is created with the right size and is 
    subsequently filled in a second loop over all subdirectories.
    This step includes preproecessing. Trajectories are checked to be finished
    and the initiation phase is discarded.
    The final data as well as the hyperparameters for every trajectory and the 
    list of features is save to `out_dir`.
    
    Keyword Arguments:
        root_dir {str}:
            Directory of the data 
        ligands {list}:
            List of ligands to load
        out_dir {str}:
            Directory where to save the results
        data_folder {str}:
            Name of the data folder (default: 'IFP-NEW')
        interaction_filename {str}:
            Name of the interaction file which stores the data  
            (default: {"full-interaction-profile.tsv"})
        initiation_threshold {int}:
            For every trajectory we discard the initiation phase. Once the 
            trajectory differs by more than `initiation_threshold` 
            /features/ % of features                                        ----- changed 25.12.2018 DK 
            from the first phase the initiation phase is considered finished; 
            everything before that will be discarded.            
            (default: {2})
        print_missing_data {bool}:
            Whether to print the names of empty folder 
            (default: {False})

    Returns:
        Nothing
    """

    # Ligands with incorrect numbering schemes
    wrong_numbering = ['2062442', 'COMP-3', 'COMP-5', 'VER-4', 'VER-5',
                    '2018086', '2036332', 'COMP-4', '2209573','4c','4d','4e','4f','4c']

    # set root_dir
    if not root_dir:
        root_dir = os.getcwd()

    # set out_dir
    if not out_dir:
        out_dir = os.getcwd()

    # set data_folder
    if not data_folder:
        data_folder = 'IFP-NEW'

    # Create list of leaf directories (add all ligands together)
    print('Start creating a list of all folder involved')
    dir_list = []
    for ligand in ligands:
        dir_list += [folder for folder, subfolders, _ in \
                os.walk(os.path.join(root_dir, ligand)) if not subfolders]
    dir_list = sorted(dir_list)

    # nr of ligands
    nr_ligands = len(ligands)

    ########################################
    # create a list of all features
    # (because the order of features differs between trajectories)
    ########################################    
    features = []
    print('Start creating a list of features')
    for i, d in enumerate(dir_list):
        current_d = d.split('/')
        start = [i for i,x in enumerate(current_d) if x==data_folder][0]
        current_d = current_d[start+1:]

        if len(list(filter(lambda x:interaction_filename in \
                                                x, os.listdir(d)))) != 0:
            # find the right file
            interaction_file = list(filter(lambda x:interaction_filename in \
                                                        x, os.listdir(d)))[0]

            # read features from header and add in list  
            with open(os.path.join(d, interaction_file)) as f:
                reader = csv.reader(f, delimiter = '\t')
                header = next(reader)

            # correct wrong numbering
            if current_d[0] in wrong_numbering:
                header = [correct_single_numbering(feature) for \
                                                    feature in header]
            
            # add features to features list
            for feature in header:
                if not feature in features: 
                    features.append(feature)

        else:
            if print_missing_data:
                print('Missing data in folder: {}'.format(d))

       
    features = np.array(features)

    ##########################################
    # Actual data loading
    ##########################################
    
    # Create empty numpy array
    nr_features = len(features)
    data = np.zeros((0, nr_features),dtype=np.int8)

    print('\nStart loading data...')
    print('\nTotal nr of trajectories: {}'.format(len(dir_list)))
    current_ligand = ''
    ligands_loaded = 0
    current_datapoint = 1
    hyperparameters = [[], [], [], []]

    # Loop through all trajectories
    for i, d in enumerate(dir_list):
        current_d = d.split('/')
        start = [i for i,x in enumerate(current_d) if x==data_folder][0]
        current_d = current_d[start+1:]

        ##########################################
        # Spot unfinished trajectories
        ##########################################                
        # load the final distance to see whether this trajectory was finished
        if len(list(filter(lambda x:'lig_center.dat' in \
                                                    x, os.listdir(d)))) != 0:
            # find the right file
            interaction_file = list(filter(lambda x:'lig_center.dat' in \
                                                        x, os.listdir(d)))[0]

            current_final_distance = np.genfromtxt(os.path.join(d, 
                                                            interaction_file), 
                                                    delimiter=' ')[-1, 1]
            if current_final_distance < 15:
                print("final distance = ",d,interaction_file,current_final_distance)
                continue
        else:
            continue

        # Sometimes the above is not enough so we have 
        # to check this second file
  #      if len(list(filter(lambda x:'egress-contact-profile.tsv' in \
  #                                                  x, os.listdir(d)))) != 0:
  #          # find the right file
  #          print("something wrong with the distance file in", d)
  #          interaction_file = list(filter(lambda x:'egress-contact-profile.tsv' in \
  #                                                      x, os.listdir(d)))[0]

   #         current_final_distance = np.genfromtxt(os.path.join(d, 
   #                                                         interaction_file), 
   #                                                 delimiter='\t')[-1, 0]
   #         if current_final_distance < 30:
   #             continue
   #     else:
   #         continue
        ##########################################
        
        if len(list(filter(lambda x:interaction_filename in \
                                                    x, os.listdir(d)))) != 0:
            # find current file and load data
            interaction_file = list(filter(lambda x:interaction_filename in \
                                                        x, os.listdir(d)))[0]        
            current_data = np.genfromtxt(os.path.join(d, interaction_file), 
                                        delimiter='\t')

            # delete header from current data
            current_data = current_data[1:, :]
            
            # create dummy_data to insert later (with the right dimensions)
            new_data = np.zeros((current_data.shape[0], nr_features),dtype = np.int8)

            # open current header 
            # (so we ne know the features for this trajectory)
            with open(os.path.join(d, interaction_file)) as f:
                reader = csv.reader(f, delimiter = '\t')
                header = next(reader)

            # correct wrong numbering   
            if current_d[0] in wrong_numbering:
                header = [correct_single_numbering(feature) for \
                                                    feature in header]

            # Fill new data with the current data.
            # Make sure the order of the features is 
            # the same for all trajectories         
            current_feature_index = [np.where(features == feature)[0][0] for \
                                                            feature in header]
            new_data[:, current_feature_index] = current_data
            
            ##########################################
            # Preprocessing: delete initiation phase
            ##########################################

            if initiation_threshold:
                # calc difference to first frame
                diffs = np.sum(np.abs(new_data - np.tile(new_data[0, :], 
                                                    (new_data.shape[0], 1))), 
                                                    axis = 1)
                if(initiation_threshold > 5):
                   if(len(np.nonzero(new_data[0, :])[0]) > 5):
                                diffsP = diffs/len(np.nonzero(new_data[0, :])[0])
                                starting_pos = np.where(diffsP*100.0 < initiation_threshold)[0][-1]
                   else:
                                starting_pos = np.where(diffs <= 1)[0][-1]
                                print("!!!only ",len(np.nonzero(new_data[0, :])[0])," contacts in starting snapshot found")
                else:
                	starting_pos = np.where(diffs <= initiation_threshold)[0][-1]
          #      print(diffs[starting_pos],"  ",diffs[starting_pos+1])
                #print(starting_pos,len(diffs),diffs[1])
                # select only data after initiation
                new_data = new_data[starting_pos:, :]

            ##########################################
            # Add new data to the existing data
            ##########################################
            data = np.concatenate([data, new_data])

        else:
            continue


        ##########################################
        # Store hyperparameters
        ##########################################
        for j, param in enumerate(d.split('/')[-3:]):
            hyperparameters[j].append(param)
        hyperparameters[3].append(new_data.shape[0])


        # print at the start of every new ligand
        if hyperparameters[0][-1] != current_ligand:
            current_ligand = hyperparameters[0][-1]
            ligands_loaded += 1
            current_ligand_length = len(list(filter(lambda x:current_ligand in \
                                                                x, dir_list)))
            print('Loading ligand {} (with {} trajectories) \t[{}/{}]'.format(
                    current_ligand, current_ligand_length, 
                    ligands_loaded, nr_ligands))
        
    ##########################################
    # Saving
    ##########################################
    print('\nSaving result to file: `data.npy`, \
        `hyperparameters.npy` and `features.npy`')
    print('data has dimensions: {}'.format(data.shape))
    np.save(os.path.join(out_dir, 'data.npy'), data)
    np.save(os.path.join(out_dir, 'hyperparameters.npy'), 
            np.array(hyperparameters).T)
    np.save(os.path.join(out_dir, 'features.npy'), features)
