import numpy as np
import pandas as pd
import re
from sklearn.preprocessing import StandardScaler, MinMaxScaler


#############################
# Data manipulation functions
#############################

def drop_empty_features(data):
    '''Return data (pandas.DataFrame) without empty columns.'''
    
    for feature in data.columns:
        if data[feature].sum() == 0:
            data.drop([feature], axis = 1, inplace = True)
    
    return data

def drop_empty_lines(data):
    '''Return data (pandas.DataFrame) without empty rows.'''

    if type(data) == pd.DataFrame:
        for line in data.iterrows():
            if line[1].sum() == 0:
                data.drop(line[0], axis = 0, inplace = True)
                
    return data

###################
# Metadata handlers
###################

def load_metadata(filename, 
                ligands = None, 
                features = None, 
                sheet_name = 0):
    '''
    Load metadata from file specified in `filename` (should be an excel 
    spreadsheet). Ligands and features can be specified.
    '''
    metadata = pd.read_excel(filename, sheet_name = sheet_name)
    metadata['filename'] = metadata['filename'].astype('str')
    metadata.index = metadata['filename']
    
    for d in pd.unique(ligands):
        if(True in metadata.index.str.contains(d)) == False:
            print ("ERROR :",d, "directory is not present in the metadata list, check or re-generate the database data.npy")
            
    if isinstance(ligands,np.ndarray):
        metadata = metadata.loc[ligands]

    if isinstance(features,list):
        metadata = metadata[features]

    return metadata


def create_frames(ligands, 
                    data, 
                    source = 'data/frames_per_trajectory.csv'):
    """
    Create DataFrame which contains the number of frames (called 'frames').
    The number of frames is a measurement of how long a simulation took.
    The different scalers are used to scale the number of frames so they are
    more easily used in machine learning methods.
    
    Arguments:
        ligands {list of str}:
            List of ligands to load
        data {pandas.DataFrame}:
            DataFrame of the ligands to load
    
    Keyword Arguments:
        source {str}:
            Filename for the metadata 
            (default: {'data/frames_per_trajectory.csv'})
    
    Returns:
        frames {pandas.DataFrame}:
            DataFrame containing the frames as well as scaled frames
        outlier_index:
            Index of entries which are not outliers
        frames_ind:
            Index sorted after size
        standard_scaler:
            fitted sklearn scaler object (this can be used for future 
            transformations and back-transformations)
        min_max_scaler
            fitted sklearn scaler object (this can be used for future 
            transformations and back-transformations)
    """

    # Load csv file
    frames = pd.read_csv(source)
    frames = frames.drop(frames.columns[0], axis = 1)
    frames = frames.loc[frames['ligand'].isin(ligands)]
    
    # create index
    frames.index = frames['ligand'].values + ':' + frames['replica'].values + \
                        ':' + frames['trajectory'].values.astype('str')
    frames.index = [re.sub('\$', '', entry) for entry in frames.index]

    # select only trajectories we have in the data
    frames = frames.loc[pd.unique(data.index.get_level_values(0))]

    # clear nan entries
    nan_index = pd.notna(frames['Frames'])
    frames.loc[~nan_index, 'Frames'] = 0

    frames.drop(['replica', 'trajectory'], axis = 1, inplace = True)
    
    # cutoff with quantile method
    Q1 = frames['Frames'].quantile(0.25)
    Q3 = frames['Frames'].quantile(0.75)
    IQR = Q3 - Q1

    outlier_index = np.logical_and(Q3 + 1.5 * IQR >= frames['Frames'], 
                    Q1 - 1.5 * IQR <= frames['Frames'])

    frames_unscaled = frames['Frames'].astype('float').values.reshape(-1, 1)
    
    # scale frames using StandardScaler (mean = 0, std = 1)
    standard_scaler = StandardScaler()
    standard_scaler.fit(frames_unscaled)
    frames_scaled = standard_scaler.transform(frames_unscaled)

    frames['StandardScaler'] = frames_scaled

    # scale frames using MinMaxScaler (all values in [0, 1])
    min_max_scaler = MinMaxScaler()
    min_max_scaler.fit(frames_unscaled)
    frames_scaled = min_max_scaler.transform(frames_unscaled)

    frames['MinMaxScaler'] = frames_scaled

    # sort w.r.t mean
    frames_ind = np.argsort(frames.loc[outlier_index].groupby('ligand').mean()['Frames'])
    
    return frames, outlier_index, \
            frames_ind, standard_scaler, min_max_scaler


def create_ligand_lists(data):
    """Given a pandas DataFrame (`data`) which was created with the function 
    `DataFrame_from_full_data` this function will different lists detailing 
    the occurrences of ligands.
    
    Trajectory-wise:
    ligands_list: List of all ligands (len = number of trajectories)
    ligand_len: How often that ligand appears
    
    Snapshot-wise:
    ligands_list_trajectory: List of all ligands (len = number of snapshots)
    ligand_len_trajectory: How often that ligand appears
    
    Arguments:
        data {pandas.DataFrame}:
            DataFrame containing the data with the trajectories as rows
            and the features as columns.
    
    Returns:
        ligands_list:
            Lists the ligand for every trajectory.
        ligand_len:
            Lists the length of every ligand for every trajectory.
        ligands_list_trajectory:
            Lists the ligand for every snapshot.
        ligand_len_trajectory:
            Lists the length of every ligand for every snapshot.
    """

    # ligand lists on trajectory level
    ligands_list = np.array([entry.split(':')[0] for \
                    entry in list(data.index.get_level_values(0).unique())])

    ligand_len = [len(np.where(ligands_list == ligand)[0]) for \
                                            ligand in pd.unique(ligands_list)]

    # ligand lists on snapshot level
    ligands_list_trajectory = np.array([entry.split(':')[0] for \
                            entry in list(data.index.get_level_values(0))])
    ligand_len_trajectory = [len(np.where(ligands_list_trajectory == ligand)[0]) for \
                                ligand in pd.unique(ligands_list_trajectory)]
    
    return ligands_list, ligand_len, ligands_list_trajectory, \
            ligand_len_trajectory